package java.util.stream;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.function.IntFunction;
import java.util.Spliterator;
import java.util.Objects;
import java.util.Comparator;

final class SortedOps
{
    private SortedOps() {
    }
    
    static <T> Stream<T> makeRef(AbstractPipeline<?, T, ?> abstractPipeline) {
        return new SortedOps.OfRef(abstractPipeline);
    }
    
    static <T> Stream<T> makeRef(AbstractPipeline<?, T, ?> abstractPipeline, Comparator<? super T> comparator) {
        return new SortedOps.OfRef(abstractPipeline, comparator);
    }
    
    static <T> IntStream makeInt(AbstractPipeline<?, Integer, ?> abstractPipeline) {
        return new OfInt(abstractPipeline);
    }
    
    static <T> LongStream makeLong(AbstractPipeline<?, Long, ?> abstractPipeline) {
        return new OfLong(abstractPipeline);
    }
    
    static <T> DoubleStream makeDouble(AbstractPipeline<?, Double, ?> abstractPipeline) {
        return new OfDouble(abstractPipeline);
    }
    
    private static final class OfRef<T> extends ReferencePipeline.StatefulOp<T, T>
    {
        private final boolean isNaturalSort;
        private final Comparator<? super T> comparator;
        
        OfRef(AbstractPipeline<?, T, ?> abstractPipeline) {
            super(abstractPipeline, StreamShape.REFERENCE, StreamOpFlag.IS_ORDERED | StreamOpFlag.IS_SORTED);
            this.isNaturalSort = true;
            this.comparator = Comparator.naturalOrder();
        }
        
        OfRef(AbstractPipeline<?, T, ?> abstractPipeline, Comparator<? super T> obj) {
            super(abstractPipeline, StreamShape.REFERENCE, StreamOpFlag.IS_ORDERED | StreamOpFlag.NOT_SORTED);
            this.isNaturalSort = false;
            this.comparator = (Comparator<? super T>)Objects.requireNonNull(obj);
        }
        
        public Sink<T> opWrapSink(int n, Sink<T> obj) {
            Objects.requireNonNull(obj);
            if (StreamOpFlag.SORTED.isKnown(n) && this.isNaturalSort) {
                return obj;
            }
            if (StreamOpFlag.SIZED.isKnown(n)) {
                return new SortedOps.SizedRefSortingSink(obj, this.comparator);
            }
            return new SortedOps.RefSortingSink(obj, this.comparator);
        }
        
        public <P_IN> Node<T> opEvaluateParallel(PipelineHelper<T> pipelineHelper, Spliterator<P_IN> spliterator, IntFunction<T[]> intFunction) {
            if (StreamOpFlag.SORTED.isKnown(pipelineHelper.getStreamAndOpFlags()) && this.isNaturalSort) {
                return pipelineHelper.evaluate(spliterator, false, intFunction);
            }
            Object[] array = pipelineHelper.evaluate(spliterator, true, intFunction).asArray(intFunction);
            Arrays.parallelSort(array, this.comparator);
            return Nodes.node(array);
        }
    }
    
    private static final class OfInt extends IntPipeline.StatefulOp<Integer>
    {
        OfInt(AbstractPipeline<?, Integer, ?> abstractPipeline) {
            super(abstractPipeline, StreamShape.INT_VALUE, StreamOpFlag.IS_ORDERED | StreamOpFlag.IS_SORTED);
        }
        
        public Sink<Integer> opWrapSink(int n, Sink<Integer> obj) {
            Objects.requireNonNull(obj);
            if (StreamOpFlag.SORTED.isKnown(n)) {
                return obj;
            }
            if (StreamOpFlag.SIZED.isKnown(n)) {
                return new SortedOps.SizedIntSortingSink(obj);
            }
            return new SortedOps.IntSortingSink(obj);
        }
        
        public <P_IN> Node<Integer> opEvaluateParallel(PipelineHelper<Integer> pipelineHelper, Spliterator<P_IN> spliterator, IntFunction<Integer[]> intFunction) {
            if (StreamOpFlag.SORTED.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                return pipelineHelper.evaluate(spliterator, false, intFunction);
            }
            int[] a = (int[])((Node.OfInt)pipelineHelper.evaluate(spliterator, true, intFunction)).asPrimitiveArray();
            Arrays.parallelSort(a);
            return Nodes.node(a);
        }
    }
    
    private static final class OfLong extends LongPipeline.StatefulOp<Long>
    {
        OfLong(AbstractPipeline<?, Long, ?> abstractPipeline) {
            super(abstractPipeline, StreamShape.LONG_VALUE, StreamOpFlag.IS_ORDERED | StreamOpFlag.IS_SORTED);
        }
        
        public Sink<Long> opWrapSink(int n, Sink<Long> obj) {
            Objects.requireNonNull(obj);
            if (StreamOpFlag.SORTED.isKnown(n)) {
                return obj;
            }
            if (StreamOpFlag.SIZED.isKnown(n)) {
                return new SortedOps.SizedLongSortingSink(obj);
            }
            return new SortedOps.LongSortingSink(obj);
        }
        
        public <P_IN> Node<Long> opEvaluateParallel(PipelineHelper<Long> pipelineHelper, Spliterator<P_IN> spliterator, IntFunction<Long[]> intFunction) {
            if (StreamOpFlag.SORTED.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                return pipelineHelper.evaluate(spliterator, false, intFunction);
            }
            long[] a = (long[])((Node.OfLong)pipelineHelper.evaluate(spliterator, true, intFunction)).asPrimitiveArray();
            Arrays.parallelSort(a);
            return Nodes.node(a);
        }
    }
    
    private static final class OfDouble extends DoublePipeline.StatefulOp<Double>
    {
        OfDouble(AbstractPipeline<?, Double, ?> abstractPipeline) {
            super(abstractPipeline, StreamShape.DOUBLE_VALUE, StreamOpFlag.IS_ORDERED | StreamOpFlag.IS_SORTED);
        }
        
        public Sink<Double> opWrapSink(int n, Sink<Double> obj) {
            Objects.requireNonNull(obj);
            if (StreamOpFlag.SORTED.isKnown(n)) {
                return obj;
            }
            if (StreamOpFlag.SIZED.isKnown(n)) {
                return new SortedOps.SizedDoubleSortingSink(obj);
            }
            return new SortedOps.DoubleSortingSink(obj);
        }
        
        public <P_IN> Node<Double> opEvaluateParallel(PipelineHelper<Double> pipelineHelper, Spliterator<P_IN> spliterator, IntFunction<Double[]> intFunction) {
            if (StreamOpFlag.SORTED.isKnown(pipelineHelper.getStreamAndOpFlags())) {
                return pipelineHelper.evaluate(spliterator, false, intFunction);
            }
            double[] a = (double[])((Node.OfDouble)pipelineHelper.evaluate(spliterator, true, intFunction)).asPrimitiveArray();
            Arrays.parallelSort(a);
            return Nodes.node(a);
        }
    }
    
    private abstract static class AbstractRefSortingSink<T> extends Sink.ChainedReference<T, T>
    {
        protected final Comparator<? super T> comparator;
        protected boolean cancellationWasRequested;
        
        AbstractRefSortingSink(Sink<? super T> sink, Comparator<? super T> comparator) {
            super(sink);
            this.comparator = comparator;
        }
        
        @Override
        public final boolean cancellationRequested() {
            this.cancellationWasRequested = true;
            return false;
        }
    }
    
    private static final class SizedRefSortingSink<T> extends SortedOps.AbstractRefSortingSink<T>
    {
        private T[] array;
        private int offset;
        
        SizedRefSortingSink(Sink<? super T> sink, Comparator<? super T> comparator) {
            super(sink, comparator);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.array = (Object[])new Object[(int)n];
            return;
        }
        
        @Override
        public void end() {
            Arrays.sort(this.array, 0, this.offset, this.comparator);
            this.downstream.begin((long)this.offset);
            if (!this.cancellationWasRequested) {
                for (int i = 0; i < this.offset; ++i) {
                    this.downstream.accept(this.array[i]);
                }
            }
            else {
                for (int n = 0; n < this.offset && !this.downstream.cancellationRequested(); ++n) {
                    this.downstream.accept(this.array[n]);
                }
            }
            this.downstream.end();
            this.array = null;
        }
        
        @Override
        public void accept(T t) {
            this.array[this.offset++] = t;
        }
    }
    
    private static final class RefSortingSink<T> extends SortedOps.AbstractRefSortingSink<T>
    {
        private ArrayList<T> list;
        
        RefSortingSink(Sink<? super T> sink, Comparator<? super T> comparator) {
            super(sink, comparator);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.list = ((n >= 0L) ? new ArrayList((int)n) : new ArrayList());
            return;
        }
        
        @Override
        public void end() {
            this.list.sort(this.comparator);
            this.downstream.begin((long)this.list.size());
            if (!this.cancellationWasRequested) {
                this.list.forEach(this.downstream::accept);
            }
            else {
                for (Object next : this.list) {
                    if (this.downstream.cancellationRequested()) {
                        break;
                    }
                    this.downstream.accept(next);
                }
            }
            this.downstream.end();
            this.list = null;
        }
        
        @Override
        public void accept(T e) {
            this.list.add(e);
        }
    }
    
    private abstract static class AbstractIntSortingSink extends Sink.ChainedInt<Integer>
    {
        protected boolean cancellationWasRequested;
        
        AbstractIntSortingSink(Sink<? super Integer> sink) {
            super(sink);
        }
        
        @Override
        public final boolean cancellationRequested() {
            this.cancellationWasRequested = true;
            return false;
        }
    }
    
    private static final class SizedIntSortingSink extends SortedOps.AbstractIntSortingSink
    {
        private int[] array;
        private int offset;
        
        SizedIntSortingSink(Sink<? super Integer> sink) {
            super(sink);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.array = new int[(int)n];
            return;
        }
        
        @Override
        public void end() {
            Arrays.sort(this.array, 0, this.offset);
            this.downstream.begin((long)this.offset);
            if (!this.cancellationWasRequested) {
                for (int i = 0; i < this.offset; ++i) {
                    this.downstream.accept(this.array[i]);
                }
            }
            else {
                for (int n = 0; n < this.offset && !this.downstream.cancellationRequested(); ++n) {
                    this.downstream.accept(this.array[n]);
                }
            }
            this.downstream.end();
            this.array = null;
        }
        
        @Override
        public void accept(int n) {
            this.array[this.offset++] = n;
        }
    }
    
    private static final class IntSortingSink extends SortedOps.AbstractIntSortingSink
    {
        private SpinedBuffer.OfInt b;
        
        IntSortingSink(Sink<? super Integer> sink) {
            super(sink);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.b = ((n > 0L) ? new SpinedBuffer.OfInt((int)n) : new SpinedBuffer.OfInt());
            return;
        }
        
        @Override
        public void end() {
            int[] a = (int[])this.b.asPrimitiveArray();
            Arrays.sort(a);
            this.downstream.begin((long)a.length);
            if (!this.cancellationWasRequested) {
                for (int length2 = a.length, j = 0; j < length2; ++j) {
                    this.downstream.accept(a[j]);
                }
            }
            else {
                int length = a.length;
                int i = 0;
                while (i < length) {
                    int n = a[i];
                    if (this.downstream.cancellationRequested()) {
                        break;
                    }
                    this.downstream.accept(n);
                    ++i;
                }
            }
            this.downstream.end();
        }
        
        @Override
        public void accept(int n) {
            this.b.accept(n);
        }
    }
    
    private abstract static class AbstractLongSortingSink extends Sink.ChainedLong<Long>
    {
        protected boolean cancellationWasRequested;
        
        AbstractLongSortingSink(Sink<? super Long> sink) {
            super(sink);
        }
        
        @Override
        public final boolean cancellationRequested() {
            this.cancellationWasRequested = true;
            return false;
        }
    }
    
    private static final class SizedLongSortingSink extends SortedOps.AbstractLongSortingSink
    {
        private long[] array;
        private int offset;
        
        SizedLongSortingSink(Sink<? super Long> sink) {
            super(sink);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.array = new long[(int)n];
            return;
        }
        
        @Override
        public void end() {
            Arrays.sort(this.array, 0, this.offset);
            this.downstream.begin((long)this.offset);
            if (!this.cancellationWasRequested) {
                for (int i = 0; i < this.offset; ++i) {
                    this.downstream.accept(this.array[i]);
                }
            }
            else {
                for (int n = 0; n < this.offset && !this.downstream.cancellationRequested(); ++n) {
                    this.downstream.accept(this.array[n]);
                }
            }
            this.downstream.end();
            this.array = null;
        }
        
        @Override
        public void accept(long n) {
            this.array[this.offset++] = n;
        }
    }
    
    private static final class LongSortingSink extends SortedOps.AbstractLongSortingSink
    {
        private SpinedBuffer.OfLong b;
        
        LongSortingSink(Sink<? super Long> sink) {
            super(sink);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.b = ((n > 0L) ? new SpinedBuffer.OfLong((int)n) : new SpinedBuffer.OfLong());
            return;
        }
        
        @Override
        public void end() {
            long[] a = (long[])this.b.asPrimitiveArray();
            Arrays.sort(a);
            this.downstream.begin((long)a.length);
            if (!this.cancellationWasRequested) {
                for (int length2 = a.length, j = 0; j < length2; ++j) {
                    this.downstream.accept(a[j]);
                }
            }
            else {
                int length = a.length;
                int i = 0;
                while (i < length) {
                    long n = a[i];
                    if (this.downstream.cancellationRequested()) {
                        break;
                    }
                    this.downstream.accept(n);
                    ++i;
                }
            }
            this.downstream.end();
        }
        
        @Override
        public void accept(long n) {
            this.b.accept(n);
        }
    }
    
    private abstract static class AbstractDoubleSortingSink extends Sink.ChainedDouble<Double>
    {
        protected boolean cancellationWasRequested;
        
        AbstractDoubleSortingSink(Sink<? super Double> sink) {
            super(sink);
        }
        
        @Override
        public final boolean cancellationRequested() {
            this.cancellationWasRequested = true;
            return false;
        }
    }
    
    private static final class SizedDoubleSortingSink extends SortedOps.AbstractDoubleSortingSink
    {
        private double[] array;
        private int offset;
        
        SizedDoubleSortingSink(Sink<? super Double> sink) {
            super(sink);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.array = new double[(int)n];
            return;
        }
        
        @Override
        public void end() {
            Arrays.sort(this.array, 0, this.offset);
            this.downstream.begin((long)this.offset);
            if (!this.cancellationWasRequested) {
                for (int i = 0; i < this.offset; ++i) {
                    this.downstream.accept(this.array[i]);
                }
            }
            else {
                for (int n = 0; n < this.offset && !this.downstream.cancellationRequested(); ++n) {
                    this.downstream.accept(this.array[n]);
                }
            }
            this.downstream.end();
            this.array = null;
        }
        
        @Override
        public void accept(double n) {
            this.array[this.offset++] = n;
        }
    }
    
    private static final class DoubleSortingSink extends SortedOps.AbstractDoubleSortingSink
    {
        private SpinedBuffer.OfDouble b;
        
        DoubleSortingSink(Sink<? super Double> sink) {
            super(sink);
        }
        
        @Override
        public void begin(long n) {
            if (n >= 2147483639L) {
                throw new IllegalArgumentException("Stream size exceeds max array size");
            }
            this.b = ((n > 0L) ? new SpinedBuffer.OfDouble((int)n) : new SpinedBuffer.OfDouble());
            return;
        }
        
        @Override
        public void end() {
            double[] a = (double[])this.b.asPrimitiveArray();
            Arrays.sort(a);
            this.downstream.begin((long)a.length);
            if (!this.cancellationWasRequested) {
                for (int length2 = a.length, j = 0; j < length2; ++j) {
                    this.downstream.accept(a[j]);
                }
            }
            else {
                int length = a.length;
                int i = 0;
                while (i < length) {
                    double n = a[i];
                    if (this.downstream.cancellationRequested()) {
                        break;
                    }
                    this.downstream.accept(n);
                    ++i;
                }
            }
            this.downstream.end();
        }
        
        @Override
        public void accept(double n) {
            this.b.accept(n);
        }
    }
}
